from wpimath.trajectory import TrajectoryGenerator, TrajectoryParameterizer
from wpimath.geometry import Pose2d, Transform2d
import wpimath.trajectory
import wpimath.trajectory.constraint
from wpimath.spline import CubicHermiteSpline, SplineHelper

from wpimath.geometry import Pose2d, Rotation2d, Translation2d

from wpimath.trajectory import Trajectory, TrajectoryConfig

from wpimath.trajectory.constraint import (
    MaxVelocityConstraint,
    EllipticalRegionConstraint,
    RectangularRegionConstraint,
)

import math
import typing


def getTestTrajectory(config: TrajectoryConfig) -> Trajectory:
    # 2018 cross scale auto waypoints
    sideStart = Pose2d.fromFeet(1.54, 23.23, Rotation2d.fromDegrees(180))
    crossScale = Pose2d.fromFeet(23.7, 6.8, Rotation2d.fromDegrees(-160))

    config.setReversed(True)

    vector = [
        (
            sideStart + Transform2d(Translation2d.fromFeet(-13, 0), Rotation2d())
        ).translation(),
        (
            sideStart
            + Transform2d(
                Translation2d.fromFeet(-19.5, 5.0), Rotation2d.fromDegrees(-90)
            )
        ).translation(),
    ]

    return TrajectoryGenerator.generateTrajectory(sideStart, vector, crossScale, config)


#
# EllipticalRegionConstraint tests
#


def test_elliptical_region_constraint():
    maxVelocity = 2
    config = TrajectoryConfig.fromFps(13, 13)
    maxVelConstraint = MaxVelocityConstraint.fromFps(maxVelocity)
    regionConstraint = EllipticalRegionConstraint.fromFeet(
        Translation2d.fromFeet(5, 2.5),
        10,
        5,
        Rotation2d.fromDegrees(180),
        maxVelConstraint,
    )

    config.addConstraint(regionConstraint)

    trajectory = getTestTrajectory(config)

    exceededConstraintOutsideRegion = False
    for point in trajectory.states():
        translation = point.pose.translation()
        if translation.x_feet < 10 and translation.y_feet < 5:
            assert abs(point.velocity_fps) < maxVelocity + 0.05
        elif abs(point.velocity_fps) >= maxVelocity + 0.05:
            exceededConstraintOutsideRegion = True

    assert exceededConstraintOutsideRegion


def test_elliptical_region_is_pose_in_region():
    maxVelConstraint = MaxVelocityConstraint.fromFps(2)
    regionConstraintNoRotation = EllipticalRegionConstraint.fromFeet(
        Translation2d.fromFeet(1, 1), 2, 4, Rotation2d(0), maxVelConstraint
    )

    assert not regionConstraintNoRotation.isPoseInRegion(Pose2d.fromFeet(2.1, 1, 0))


#
# RectangularRegionConstraint tests
#


def test_rectangular_region_constraint():
    maxVelocity = 2
    config = TrajectoryConfig.fromFps(13, 13)
    maxVelConstraint = MaxVelocityConstraint.fromFps(maxVelocity)
    regionConstraint = RectangularRegionConstraint(
        Translation2d.fromFeet(1, 1),
        Translation2d.fromFeet(5, 27),
        maxVelConstraint,
    )

    config.addConstraint(regionConstraint)

    trajectory = getTestTrajectory(config)

    exceededConstraintOutsideRegion = False
    for point in trajectory.states():
        if regionConstraint.isPoseInRegion(point.pose):
            assert abs(point.velocity_fps) < maxVelocity + 0.05
        elif abs(point.velocity_fps) >= maxVelocity + 0.05:
            exceededConstraintOutsideRegion = True

    assert exceededConstraintOutsideRegion


def test_rectangular_region_is_pose_in_region():
    maxVelConstraint = MaxVelocityConstraint.fromFps(2)
    regionConstraint = RectangularRegionConstraint(
        Translation2d.fromFeet(1, 1), Translation2d.fromFeet(5, 27), maxVelConstraint
    )
    assert not regionConstraint.isPoseInRegion(Pose2d())
    assert regionConstraint.isPoseInRegion(Pose2d.fromFeet(3, 14.5, Rotation2d()))


#
# TrajectoryParameterizer
#


def test_trajectory_parameterizer():
    start = Pose2d(1, 1, 0)
    end = Pose2d(2, 2, math.pi / 2)

    # generate the spline from start and end poses
    vec1, vec2 = SplineHelper.cubicControlVectorsFromWaypoints(start, [], end)
    spline = CubicHermiteSpline(vec1.x, vec2.x, vec1.y, vec2.y)

    # sample the pose and curvature along the spline
    points: typing.Tuple[Pose2d, float] = []
    for i in range(100):
        points.append(spline.getPoint(i / 100))

    trajectory = TrajectoryParameterizer.timeParameterizeTrajectory(
        points, [], 0, 0, 4, 3, False
    )
    assert trajectory is not None



from . import _init_geometry

# autogenerated by 'robotpy-build create-imports wpimath.geometry'
from ._geometry import (
    CoordinateAxis,
    CoordinateSystem,
    Pose2d,
    Pose3d,
    Quaternion,
    Rotation2d,
    Rotation3d,
    Transform2d,
    Transform3d,
    Translation2d,
    Translation3d,
    Twist2d,
    Twist3d,
)

__all__ = [
    "CoordinateAxis",
    "CoordinateSystem",
    "Pose2d",
    "Pose3d",
    "Quaternion",
    "Rotation2d",
    "Rotation3d",
    "Transform2d",
    "Transform3d",
    "Translation2d",
    "Translation3d",
    "Twist2d",
    "Twist3d",
]



# Copyright (c) FIRST and other WPILib contributors.
# Open Source Software; you can modify and/or share it under the terms of
# the WPILib BSD license file in the root directory of this project.

import math

kInchesPerFoot = 12.0
kMetersPerInch = 0.0254
kSecondsPerMinute = 60
kMillisecondsPerSecond = 1000
kKilogramsPerLb = 0.453592


def metersToFeet(meters: float) -> float:
    """Converts given meters to feet.

    :param meters: The meters to convert to feet.

    :returns: Feet converted from meters.
    """
    return metersToInches(meters) / kInchesPerFoot


def feetToMeters(feet: float) -> float:
    """Converts given feet to meters.

    :param feet: The feet to convert to meters.

    :returns: Meters converted from feet.
    """
    return inchesToMeters(feet * kInchesPerFoot)


def metersToInches(meters: float) -> float:
    """Converts given meters to inches.

    :param meters: The meters to convert to inches.

    :returns: Inches converted from meters.
    """
    return meters / kMetersPerInch


def inchesToMeters(inches: float) -> float:
    """Converts given inches to meters.

    :param inches: The inches to convert to meters.

    :returns: Meters converted from inches.
    """
    return inches * kMetersPerInch


# Converts given degrees to radians.
degreesToRadians = math.radians

# Converts given radians to degrees.
radiansToDegrees = math.degrees


def radiansToRotations(radians: float) -> float:
    """Converts given radians to rotations.

    :param radians: The radians to convert.

    :returns: rotations Converted from radians.
    """
    return radians / math.tau


def degreesToRotations(degrees: float) -> float:
    """Converts given degrees to rotations.

    :param degrees: The degrees to convert.

    :returns: rotations Converted from degrees.
    """
    return degrees / 360


def rotationsToDegrees(rotations: float) -> float:
    """Converts given rotations to degrees.

    :param rotations: The rotations to convert.

    :returns: degrees Converted from rotations.
    """
    return rotations * 360


def rotationsToRadians(rotations: float) -> float:
    """Converts given rotations to radians.

    :param rotations: The rotations to convert.

    :returns: radians Converted from rotations.
    """
    return rotations * math.tau


def rotationsPerMinuteToRadiansPerSecond(rpm: float) -> float:
    """Converts rotations per minute to radians per second.

    :param rpm: The rotations per minute to convert to radians per second.

    :returns: Radians per second converted from rotations per minute.
    """
    return (rpm / kSecondsPerMinute) * math.tau


def radiansPerSecondToRotationsPerMinute(radiansPerSecond: float) -> float:
    """Converts radians per second to rotations per minute.

    :param radiansPerSecond: The radians per second to convert to from rotations per minute.

    :returns: Rotations per minute converted from radians per second.
    """
    return (radiansPerSecond * kSecondsPerMinute) / math.tau


def millisecondsToSeconds(milliseconds: float) -> float:
    """Converts given milliseconds to seconds.

    :param milliseconds: The milliseconds to convert to seconds.

    :returns: Seconds converted from milliseconds.
    """
    return milliseconds / kMillisecondsPerSecond


def secondsToMilliseconds(seconds: float) -> float:
    """Converts given seconds to milliseconds.

    :param seconds: The seconds to convert to milliseconds.

    :returns: Milliseconds converted from seconds.
    """
    return seconds * kMillisecondsPerSecond


def kilogramsToLbs(kilograms: float) -> float:
    """Converts kilograms into lbs (pound-mass).

    :param kilograms: The kilograms to convert to lbs (pound-mass).

    :returns: Lbs (pound-mass) converted from kilograms.
    """
    return kilograms / kKilogramsPerLb


def lbsToKilograms(lbs: float) -> float:
    """Converts lbs (pound-mass) into kilograms.

    :param lbs: The lbs (pound-mass) to convert to kilograms.

    :returns: Kilograms converted from lbs (pound-mass).
    """
    return lbs * kKilogramsPerLb



from . import _init_interpolation  # noqa: F401

# autogenerated by 'robotpy-build create-imports wpimath.interpolation'
from ._interpolation import (
    TimeInterpolatableFloatBuffer,
    TimeInterpolatablePose2dBuffer,
)

__all__ = ["TimeInterpolatableFloatBuffer", "TimeInterpolatablePose2dBuffer"]
